/*
 * Nest simulation library - process definitions
 * 
 * $Header: process.h,v 2.5 88/03/31 14:49:53 dupuy Rel $
 */

/*
 * Process states are used to keep track of the status of the function(s)
 * running on a node.  The `_proc_table' of processes for the nodes in the
 * simulation is currently the only use of process states.
 * 
 * A process state is a compound structure as defined below:
 */

#include "state.h"                      /* to get local "istate" definition */

#ifndef Graph_Magic                     /* prevent redeclaration of position */
typedef struct                          /* must match the typedef in graph.h */
{
    long            x;
    long            y;
    long            z;
} position;

#endif

typedef struct process_state
{
    funcptr function;                   /* pointer to function for this node */

    flag            started;            /* true if started, false otherwise */
    flag            repeat;             /* true if node should be repeated */
    int             whyblocked;         /* cause of blocking (see below) */

    timev           unblocked;          /* time when this node could restart */
    timev           runtime;            /* virtual run time for this node */
    timev           cputime;            /* virtual cpu time for this node */
    timev           timeleft;           /* time left in this pass for node */
    int             stopcount;          /* _stop_timer count for this node */
    int             holdcount;          /* signal hold count for this node */

    istate          environment;        /* buffer for internal machine state */
} process;

/*
 * The function field is the function which is running as this process.
 * 
 * The started field indicates the state of the process's function.  It is set to
 * true when the function is called, and set to false when the function
 * returns.
 * 
 * The whyblocked field indicates the cause of blocking, if any, and also whether
 * a node is blocked at all.  It is set by the _block() function and cleared
 * (set to 0 = UNBLOCKED) each time the node is started or restarted.
 * 
 * The unblocked field is the time at which a node may be scheduled for activity.
 * This may be the arrival time of the first message which is available to
 * unblock the process if it is waiting, or the synchronization time of the
 * last pass it has completed.  The scheduler tables are sorted by this field;
 * it might be easier to have it in its own table or something. The runtime
 * field is the elapsed time for the process.  The cputime field is the
 * elapsed cpu time for the process.  The timeleft field is the amount of
 * additional cpu time the process will be allowed to use before it is swapped
 * out.  They are all "timeval" structures as defined in the file
 * <sys/time.h>, with "long" fields for both microseconds and seconds.
 * 
 * The stopcount and holdcount fields are counts of the number of times the
 * timers have been stopped, or signals have been blocked, in the process.
 * 
 * The environment field is a buffer for the machine (dependent) state, and
 * should only be used by the assembler routines that do the switching of
 * processes.  It is similar to a "jmp_buf" structure, as defined in the file
 * <setjmp.h>, but may vary from machine to machine.
 */

/*
 * The following are the codes defined for the whyblocked field.
 */

#define UNBLOCKED    0                  /* not blocked at all */
#define RECV(n_id)  (n_id)              /* blocked on receive from node n_id */
#define RECEIVE     -1                  /* blocked on receive from anyone */
#define WAIT        -2                  /* blocked for a time; ok to wake up */
#define SLEEP       -3                  /* blocked for a time; don't wake up */
#define ADVANCE     -4                  /* blocked to use up fake cpu time */
#define DEAD        -5                  /* real runtime error: bughalt */

/*
 * The following are integer codes for process states which are actually
 * represented by being in one of the various scheduler queues.  They are used
 * as arguments by functions which manipulate the scheduler table, so that a
 * function which deals with the process table can ignore scheduler types.
 */

#ifndef DoneQueue
#define DoneQueue            0          /* = started is false */
#define BlockedQueue        -1          /* = not time_nonzero(unblocked) */
#define WaitQueue            1          /* = time_after(sync_time,unblocked) */
#define NewpassWaitqueue     2          /* = time_after(unblocked,sync_time) */
#endif

/*
 * The `_proc_table', which keeps track of the state of the function
 * associated with each node, is simply an array (indexed by node id's) of
 * "process_state" structures, and is defined like this:
 */

extern process *_proc_table;            /* size of `nodes' arg to simulate() */
