/*
 * Nest simulation library - additional scheduler routines
 * 
 * $Header: sched2s.c,v 2.5 88/03/31 14:50:13 dupuy Rel $
 */

#include "nest.h"
#include "graph.h"

#include "defs.h"
#include "simulate.h"
#include "process.h"
#include "schedule.h"

/*
 * This file contains scheduling utility functions which use process table
 * information to keep the scheduler tables in time order.  If the unblocked
 * time data is ever added to the scheduler table, this can be merged with the
 * other file of scheduling utility functions.
 * 
 * Function for switching from current execution queue to (newpass) queue
 */

ident           _next_pass (squeue)
node           *squeue;
{
    ident           next_node = squeue->n_id;


    if (not time_after (_sync_time, _proc_table[next_node].unblocked))
        longjmp (_scheduler, NewPass);  /* no nodes to run this pass */

    _wait_queue = _next_waitqueue;

    /* move _next_waitqueue to the first with unblocked > sync_time */

    while (_next_waitqueue isnt nil and
            time_before (_proc_table[_next_waitqueue->n_id].unblocked,
                         _sync_time))

        _next_waitqueue = _next_waitqueue->next;

    /* break _next_waitqueue from _wait_queue there */

    if (_next_waitqueue isnt nil)
    {
        report (("split newpass waitqueue: %d\n", _next_waitqueue->n_id));

        if (_next_waitqueue->prev isnt nil)
            _next_waitqueue->prev->next = nil;

        _next_waitqueue->prev = nil;

        if (_next_waitqueue is _wait_queue)
            _wait_queue = nil;
        else
            report ((_show_sched ("wait queue", _wait_queue)));

        report ((_show_sched ("newpass waitqueue", _next_waitqueue)));
    }

    _execution_last = _next_last;
    _next_queue = _next_last = nil;

    return (next_node);
}


/*
 * Functions for inserting a blocked node onto the appropriate queue.
 */

_insert_sched (nodeid)
ident           nodeid;
{
    register node  *newnode = &(_sched_table[nodeid]);
    timev          *newblocked = &(_proc_table[nodeid].unblocked);

    report (("inserting node %d on ", nodeid));

    if (time_iszero (*newblocked))
    {                                   /* put at the front of blocked queue */
        report (("blocked queue\n"));
        _blocked_queue = _push_node (nodeid, _blocked_queue);
        report ((_show_sched ("blocked queue", _blocked_queue)));
    }
    else
    {
        if (time_before (*newblocked, _sync_time))
        {                               /* put into sorted _wait_queue */
            report ((" wait queue\n"));
            _wait_queue =
                when (_wait_queue isnt nil)
                (
                 then (_sort_sched (_wait_queue, newnode, newblocked))
                 otherwise (_waitqueue_create (newnode, _execution_last))
                );
            report ((_show_sched ("wait queue", _wait_queue)));
        }
        else
        {                               /* put into sorted _next_waitqueue */
            report ((" newpass waitqueue\n"));
            _next_waitqueue =
                when (_next_waitqueue isnt nil)
                (
                 then (_sort_sched (_next_waitqueue, newnode, newblocked))
                 otherwise (_waitqueue_create (newnode, _next_last))
                );
            report ((_show_sched ("_next_waitqueue", _next_waitqueue)));
        }
    }
}



node           *_sort_sched (squeue, newnode, newblocked)
register node  *squeue,
               *newnode;
register timev *newblocked;
{
    node           *queuehead = squeue;


    while (time_after (*newblocked, _proc_table[squeue->n_id].unblocked))
    {
        if (squeue->next isnt nil)
            squeue = squeue->next;
        else
        {
            newnode->prev = squeue;
            newnode->next = squeue->next;
            squeue->next = newnode;
            return (queuehead);
        }
    }

    newnode->prev = squeue->prev;
    newnode->next = squeue;

    if (squeue->prev isnt nil)
        squeue->prev->next = newnode;

    squeue->prev = newnode;

    if (queuehead isnt squeue)
        return (queuehead);
    else
        return (newnode);
}



/*
 * Functions for deleting a blocked node from the appropriate queue.
 */

_delete_sched (nodeid)
ident           nodeid;
{
    timev          *newblocked = &(_proc_table[nodeid].unblocked);

    report (("deleting node %d from ", nodeid));

    if (time_iszero (*newblocked))
    {                                   /* drop from blocked queue */
        report (("blocked queue\n"));
        _remove_sched (nodeid, BlockedQueue);
        report ((_show_sched ("blocked queue", _blocked_queue)));
    }
    else
    {
        if (time_before (*newblocked, _sync_time))
        {                               /* drop from _wait_queue */
            report ((" wait queue\n"));
            _remove_sched (nodeid, WaitQueue);
            report ((_show_sched ("wait queue", _wait_queue)));
        }
        else
        {                               /* drop from _next_waitqueue */
            report ((" newpass waitqueue\n"));
            _remove_sched (nodeid, NewpassWaitqueue);
            report ((_show_sched ("_next_waitqueue", _next_waitqueue)));
        }
    }
}

ident           _first_free_node ()     /* find first node with nil function */
{
    node           *squeue;


    squeue = _done_queue;
    foreach (squeue)
        if (_proc_table[squeue->n_id].function is nil)
        return (squeue->n_id);

    return (nil_id);
}


_node_update (nodedata)
grnodedat      *nodedata;
{
    reg ident       nodeid;
    process        *proc;


    nodeid = nodedata->nodeid;
    proc = &_proc_table[nodeid];

    /* nodes being deleted have nil function; they should be stopped too */

    proc->function = nodedata->function;
    proc->repeat = nodedata->repeat;

    if (proc->started is true           /* if running & get halt request */
            and nodedata->halt is true)
    {
        proc->started = false;          /* clear flag; dispose stack space */

        if (nest_destroy (nodeid) is Error)
            fatal ("sched2s: destroy: ");

        _delete_sched (nodeid);         /* remove from scheduler queue */

        _done_queue = _push_node (nodeid, _done_queue);
    }                                   /* and insert on done queue */

    if (proc->started isnt true         /* if down & get start request */
            and nodedata->start is true)
    {
        nest_proc_start (nodeid);           /* insert on newpassqueue */

        nodedata->status = RUNNING;
    }

    nodedata->halt = nodedata->start = false;
}

