/*
 * Nest simulation library - scheduler routines
 */

#include "nest.h"

#include "defs.h"
#include "simulate.h"
#include "schedule.h"

/*
 * The `_sched_table', which holds the four distinct scheduling queues that
 * make up the scheduler database, is an array (indexed by node id's) of
 * "node_queue" structures, with embedded node queues known as the execution,
 * new (or next) pass, blocked, and done queues, and is defined like this:
 */

nodeptr  _sched_table;           /* size of `nodes' arg to simulate() */

nodeptr  _done_queue;            /* nodes whose functions returned */
nodeptr  _blocked_queue;         /* nodes which have blocked on recv */

nodeptr  _execution_last;        /* last node on execution queue */

nodeptr  _wait_queue;            /* #1 waiting node on exec. queue */

nodeptr  _next_waitqueue;        /* #1 waiting node on new pass queue */

nodeptr  _next_queue;            /* #1 node for next scheduler pass */

nodeptr  _next_last;             /* last node for next scheduler pass */

/*
 * no `_execution_queue' variable exists; since _current_node is the node id
 * of the first node on the execution queue, a node pointer would be
 * redundant.
 */

/*
 * Initialization and maintenance of scheduler table
 */

_sched_init (nodes)
ident           nodes;
{
    reg ident       id;


    _sched_table = newarray (nodes + 1, node);
    _blocked_queue = nil;
    _execution_last = nil;    
    _wait_queue = nil;         
    _next_waitqueue = nil;      
    _next_queue = nil;           
    _next_last = nil;             
    _done_queue = nil;

    id = 1;
    while (id <= nodes)
    {
        _done_queue = _push_node (id, _done_queue);
        _done_queue->n_id = id++;
    }
}


_newpass_add (nodeid)
reg ident       nodeid;
{
    if (_next_queue is nil)             /* link first to _next_waitqueue */
    {
        report (("newpass_queue started with node %d\n", nodeid));
        _next_last = _next_queue = _push_node (nodeid, _next_waitqueue);
    }
    else                                /* push later ones on _next_queue */
    {
        report (("node %d added to start of newpass_queue\n", nodeid));
        _next_queue = _push_node (nodeid, _next_queue);
    }
    report ((_show_sched ("newpass_queue", _next_queue)));

}


node           *_push_node (nodeid, squeue)
ident           nodeid;
register node  *squeue;
{
    register node  *firstnode = &(_sched_table[nodeid]);

    firstnode->next = squeue;
    if (squeue isnt nil)
        squeue->prev = firstnode;
    return (firstnode);
}


_remove_sched (nodeid, state)
ident           nodeid;
int             state;
{
    register node  *dropnode = &(_sched_table[nodeid]),
                   *nextnode = dropnode->next,
                   *prevnode = dropnode->prev,
                  **queuep;


    report (("removing node %d from ", nodeid));

    if (nextnode isnt nil)
        nextnode->prev = prevnode;
    if (prevnode isnt nil)
        prevnode->next = nextnode;

    switch (state)
    {
      case DoneQueue:
        report (("done queue\n"));
        queuep = &_done_queue;
        break;

      case BlockedQueue:
        report (("blocked queue\n"));
        queuep = &_blocked_queue;
        break;

      case WaitQueue:
        report (("wait queue\n"));
        queuep = &_wait_queue;
        break;

      case NewpassWaitqueue:
        report (("newpass waitqueue\n"));
        queuep = &_next_waitqueue;
    }

    if (*queuep is dropnode)
        *queuep = nextnode;

    dropnode->next = dropnode->prev = nil;
}



ident           _get_next_node (nodeid)
ident           nodeid;
{
    register node  *curnode = &_sched_table[nodeid],
                   *nextnode = curnode->next,
                   *prevnode = curnode->prev;

    report (("get next node"));

    /* unlink current node from everything else */
    curnode->next = curnode->prev = nil;

    /* if we are at beginning of _wait_queue, advance _wait_queue */
    if (_wait_queue isnt nil and nodeid is _wait_queue->n_id)
        _wait_queue = nextnode;

    /* if we are at the end of execution_queue, set _execution_last to nil. */
    if (_execution_last isnt nil and nodeid is _execution_last->n_id)
        _execution_last = nil;

    if (prevnode isnt nil)              /* waitqueue entries inserted by send */
    {
        prevnode->next = nextnode;      /* bypass current node entry */
        if (nextnode isnt nil)
            nextnode->prev = prevnode;

        report ((": %d from wait queue\n", _wait_queue->n_id));
        return (_wait_queue->n_id);     /* return first _wait_queue entry */
    }
    else
    {
        if (nextnode isnt nil)          /* use it for next node */
        {
            nextnode->prev = nil;       /* detach it from current node */
            report ((": %d from ", nextnode->n_id));
            report ((
                     when (_execution_last is nil or _wait_queue is nextnode)
                     (
                      then ("waitqueue\n")
                      otherwise ("execution queue\n")
                      )));
            return (nextnode->n_id);
        }
        else                            /* no more nodes to run */
        {
            report ((" from next pass\n"));
            return (0);
        }
    }
}

node           *_waitqueue_create (newnode, prev)
register node  *newnode,
               *prev;
{
    newnode->prev = prev;
    newnode->next = nil;
    if (prev isnt nil)
        prev->next = newnode;
    return (newnode);
}


#ifdef debug
char           *_show_sched (string, squeue)
char           *string;
node           *squeue;
{
#define MAXINDEX    256
    static char     form[2 * MAXINDEX];
    int             index;


    (void) sprintf (form, "%s:\n\t", string);
    index = strlen (form);
    if (squeue isnt nil)
    {
        for (; squeue->next isnt nil; squeue = squeue->next)
            if (squeue->next isnt squeue and index < MAXINDEX - 25)
            {
                (void) sprintf (&form[index], "%d=>", squeue->n_id);
                index += strlen (&form[index]);
            }
            else
            {
                (void) sprintf (&form[index], "%d=>...=>", squeue->n_id);
                index += strlen (&form[index]);
                squeue = squeue->next;
                break;
            }
        (void) sprintf (&form[index], "%d\n\t", squeue->n_id);
        index += strlen (&form[index]);

        for (; squeue->prev isnt nil; squeue = squeue->prev)
            if (squeue->prev isnt squeue and index < 2 * MAXINDEX - 25)
            {
                (void) sprintf (&form[index], "%d<=", squeue->n_id);
                index += strlen (&form[index]);
            }
            else
            {
                (void) sprintf (&form[index], "%d<=...<=", squeue->n_id);
                index += strlen (&form[index]);
                squeue = squeue->next;
                break;
            }
        (void) sprintf (&form[index], "%d\n", squeue->n_id);
    }
    return (form);
}

#endif
